import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;
import DialogLayout;
import MainMenuRes;
import MainWinRes;
import OptionsRes;
import HelpAboutRes;
import MessageBoxRes;
import MessageBox;

public class EX18B extends Frame implements Runnable, SubscriptionClient
{
    static final int DEBT_PORT = 1996;
    MainMenuRes menu;
    MainWinRes mainWin;
    SubscriptionManager manager = null;
    int packetsSent = 0, packetsRecd = 0;
    Thread running = null;

    public static void main(String args[])
    {
        new EX18B("National Debt Server");
    }

    public EX18B(String caption)
    {
        super(caption);

        // Create main menu built using Resource Wizard.
        menu = new MainMenuRes(this);
        menu.CreateMenu();

        // Setup font to use.
        setFont(new Font("Dialog", Font.PLAIN, 8));
        mainWin = new MainWinRes(this);
        mainWin.CreateControls();

        // Initialize main window fields.
        mainWin.IDC_CURRENTDEBT.setText("0.00");
        mainWin.IDC_DAILYAVERAGE.setText("0.00");
        mainWin.IDC_PORTNUMBER.setText(String.valueOf(DEBT_PORT));
        mainWin.IDC_CLIENTCOUNT.setText("0");
        mainWin.IDC_DATAGRAMSENT.setText(String.valueOf(packetsSent));
        mainWin.IDC_DATAGRAMRECD.setText(String.valueOf(packetsRecd));

        // Start up the subscription manager to listen for applets.
        manager = new SubscriptionManager(this, DEBT_PORT);

        show();

        // Start up application thread to broadcast debt updates.
        running = new Thread(this);
        running.start();
    }

    public boolean handleEvent(Event event)
    {
        boolean retval = true;

        if (event.target == menu.ID_FILE_OPTIONS)
        {   // Show options dialog.
            new OptionsDlg(this, new OptionsData(mainWin.IDC_CURRENTDEBT.getText(), 
                    mainWin.IDC_DAILYAVERAGE.getText()));
        }
        else if (event.target == menu.ID_FILE_EXIT || event.id == Event.WINDOW_DESTROY)
        {   // Kill manager and broadcast threads.
            if (manager != null)
                manager.stop();
            if (running != null)
                running.stop();
            dispose();
            System.exit(0);
        }
        else if (event.target == menu.ID_HELP_ABOUT)
        {   // Display help|about.
            new HelpAboutDlg(this);
        }
        else if (event.arg instanceof OptionsData)
        {   // Update options data based on values from options dialog.
            OptionsData data = (OptionsData)event.arg;
            mainWin.IDC_CURRENTDEBT.setText(data.currentDebt);
            mainWin.IDC_DAILYAVERAGE.setText(data.dailyAvg);
        }
        else
            retval = false;

        return retval;
    }

    public void run()
    {   // Thread responsible for broadcasting debt updates to subscribing applets.
        DatagramSocket socket = null;
        while (true)
        {
            try
            {   // Broadcast every 3 seconds.
                Thread.sleep(3000);
                if (manager.getSubscriberCount() > 0)
                {   // Have subscribers so get debt, setup socket and send datagrams.
                    String debt = getCurrentDebt();
                    byte buf[] = new byte[debt.length()];
                    debt.getBytes(0, debt.length(), buf, 0);
                    socket = new DatagramSocket(DEBT_PORT);
                    for (Enumeration e = manager.subscribers(); e.hasMoreElements();)
                    {   // Send datagram to each applet.
                        InetAddress addr = (InetAddress)e.nextElement();
                        DatagramPacket packet = new DatagramPacket(buf, buf.length, addr, DEBT_PORT);
                        socket.send(packet);
                        datagramSent();
                    }
                }
            }
            catch (InterruptedException e)
                break;
            catch (Exception e)
            {
                new MessageBox(this, "Communications error.  Server stopping.", "Error");
                break;
            }
        }
        if (socket != null)
            socket.close();
    }

    public synchronized void subscriberCountChange(int count)
    {   // Called by SubscriptionManager when count changes.
        mainWin.IDC_CLIENTCOUNT.setText(String.valueOf(count));
        mainWin.IDC_CLIENTLIST.clear();
        for (Enumeration e = manager.subscribers(); e.hasMoreElements();)
            mainWin.IDC_CLIENTLIST.addItem(e.nextElement().toString());
    }

    public synchronized  void datagramSent()
    {   // Called by SubscriptionManager when a datagram is sent.
        mainWin.IDC_DATAGRAMSENT.setText(String.valueOf(++packetsSent));
    }

    public synchronized void datagramRecd()
    {   // Called by SubscriptionManager when a datagram is received.
        mainWin.IDC_DATAGRAMRECD.setText(String.valueOf(++packetsRecd));
    }

    protected String getCurrentDebt()
    {   // Calculate current debt based on previous day's balance plus
        // amount of increase up to current time.  Note: assumes that
        // debt will always increase (pretty reliable).
        String retval;
        try
        {
            Date now = new Date();
            double debt = Double.valueOf(mainWin.IDC_CURRENTDEBT.
                    getText()).doubleValue();
            double avg = Double.valueOf(mainWin.IDC_DAILYAVERAGE.getText()).doubleValue();
            double perSecond = avg / 86400;
            int seconds = (now.getHours() * 3600) + (now.getMinutes() * 60) + now.getSeconds();
            double currDebt = (perSecond * seconds) + debt;
            retval = formatBigCurrency(currDebt);
        }
        catch (NumberFormatException e) 
            retval = "$ 0.00";
        
        return retval;
    }

    protected String formatBigCurrency(double value)
    {   // Formats a big double as a currency string since 
        // toString() returns exponential notation.
        double power = 1000000000000.00d;
        String s = "$ ";

        while ((long)value > 0)
        {
            double work = value / power;
            if ((long)work > 0)
            {
                if (work < 100 && s.length() > 2)
                    s += "0";
                if (work < 10 && s.length() > 2)
                    s += "0";
                int piece = (int)work;
                s += piece;
                value -= (piece * power);
                if ((long)value > 0)
                    s += ",";
            }
            else if (s.length() > 2)
                s += "000,";
            power /= 1000f;
        }
        s += ".";
        int cents = (int)(value * 100);
        if (cents < 10)
            s += "0";
        s += cents;

        return s;
    }
}

// Interface used by SubscriptionManager to communicate with client object.
interface SubscriptionClient
{
    public void subscriberCountChange(int count);
    public void datagramSent();
    public void datagramRecd();
}

// Monitors given port for subscription requests and cancellations.
class SubscriptionManager extends Thread
{
    SubscriptionClient client;  // Client object for subscription services.
    Vector subscribers;         // List of subscribers.
    DatagramSocket socket;
    int port = 0;
    static final String SUB_CMD = "SUBSCRIBE";
    static final String UNSUB_CMD = "UNSUBSCRIBE";

    public SubscriptionManager(SubscriptionClient client, int port)
    {
        this.client = client;
        this.port = port;
        subscribers = new Vector();
        start();
    }

    public int getSubscriberCount()
    {
        return subscribers.size();
    }

    public Enumeration subscribers()
    {
        return subscribers.elements();
    }

    protected void addSubscriber(InetAddress newAddr)
    {   // Add a subscriber if not already in list.
        if (!subscribers.contains(newAddr))
        {
            subscribers.addElement(newAddr);
            // Notify client object.
            client.subscriberCountChange(subscribers.size());
        }
    }

    protected void removeSubscriber(InetAddress existAddr)
    {   // Remove a subscriber if in list.
        int idx = subscribers.indexOf(existAddr);
        if (idx != -1)
        {
            subscribers.removeElementAt(idx);
            // Notify client object.
            client.subscriberCountChange(subscribers.size());
        }
    }

    public void run()
    {   // Process subscription requests and cancellations. 
        try
        {
            DatagramPacket recv, send;
            socket = new DatagramSocket(port);
            String cmd;

            while (true)
            {
                recv = new DatagramPacket(new byte[128], 128);
                socket.receive(recv);
                client.datagramRecd();
                cmd = new String(recv.getData(), 0, 0, recv.getLength());
                if (cmd.equals(SUB_CMD))
                    addSubscriber(recv.getAddress());
                else if (cmd.equals(UNSUB_CMD))
                    removeSubscriber(recv.getAddress());
                // Return command received as a form of confirmation so 
                // applet knows we're here.
                send = new DatagramPacket(recv.getData(), recv.getLength(),
                        recv.getAddress(), port);
                socket.send(send);
                client.datagramSent();
            }
        }
        catch (SocketException se)
            System.err.println("Socket error: " + se);
        catch (IOException e)
            System.err.println("IO Error: " + e);
    }
}

// Used to exchange data between dialog and parent window.
class OptionsData
{
    public String currentDebt;
    public String dailyAvg;

    public OptionsData()
    {
        this("0.00", "0.00");
    }

    public OptionsData(String current, String avg)
    {
        currentDebt = current;
        dailyAvg = avg;
    }
}

// Options dialog used to set current debt and average debt.
class OptionsDlg extends Dialog
{
    OptionsRes resource = null;

    public OptionsDlg(Frame parent, OptionsData data)
    {
        super(parent, "Debt Options", true);
        resource = new OptionsRes(this);
        resource.CreateControls();
        resource.IDC_CURRENTDEBT.setText(data.currentDebt);
        resource.IDC_DAILYAVERAGE.setText(data.dailyAvg);
        show();
    }

    public boolean handleEvent(Event event)
    {
        boolean retval = false;

        if (event.target instanceof Button)
        {
            retval = true;
            if (event.target == resource.IDOK)
            {
                String sDebt = resource.IDC_CURRENTDEBT.getText();
                String sAvg = resource.IDC_DAILYAVERAGE.getText();
                try
                {
                    Double.valueOf(sDebt);
                    Double.valueOf(sAvg);
                    OptionsData data = new OptionsData(sDebt, sAvg);
                    Event evt = new Event(this, 0, data);
                    getParent().postEvent(evt);
                }
                catch (NumberFormatException e)
                {
                    new MessageBox((Frame)getParent(), "Invalid number format.", "Error");
                    retval = false;
                }
            }
            if (retval)
                dispose();
        }
        return retval;
    }

}

// Help|About dialog for application.
class HelpAboutDlg extends Dialog
{
    HelpAboutRes resource = null;

    public HelpAboutDlg(Frame parent)
    {
        super(parent, "About Datagram Server", true);
        resource = new HelpAboutRes(this);
        resource.CreateControls();
        show();
    }

    public boolean handleEvent(Event event)
    {
        boolean retval = false;

        if (event.target == resource.IDOK)
        {
            hide();
            retval = true;
        }
        return retval;
    }

}

